#!/usr/bin/env python2
# -*- coding: utf-8 -*-
# File: simulator.py
# Author: Yuxin Wu <ppwwyyxxc@gmail.com>

import multiprocessing as mp
import time
import threading
import weakref
from abc import abstractmethod, ABCMeta
from collections import defaultdict, namedtuple
import numpy as np
import six
from six.moves import queue

from ..utils.timer import *
from ..utils.serialize import *
from ..utils.concurrency import *

__all__ = ['SimulatorProcess', 'SimulatorMaster',
        'StateExchangeSimulatorProcess', 'SimulatorProcessSharedWeight']

try:
    import zmq
except ImportError:
    logger.warn("Error in 'import zmq'. RL simulator won't be available.")
    __all__ = []

class SimulatorProcessBase(mp.Process):
    __metaclass__ = ABCMeta

    def __init__(self, idx):
        super(SimulatorProcessBase, self).__init__()
        self.idx = int(idx)
        self.identity = u'simulator-{}'.format(self.idx).encode('utf-8')

    @abstractmethod
    def _build_player(self):
        pass


class StateExchangeSimulatorProcess(SimulatorProcessBase):
    """
    A process that simulates a player and communicates to master to
    send states and receive the next action
    """
    __metaclass__ = ABCMeta

    def __init__(self, idx, pipe_c2s, pipe_s2c):
        """
        :param idx: idx of this process
        """
        super(StateExchangeSimulatorProcess, self).__init__(idx)
        self.c2s = pipe_c2s
        self.s2c = pipe_s2c

    def run(self):
        player = self._build_player()
        context = zmq.Context()
        c2s_socket = context.socket(zmq.PUSH)
        c2s_socket.setsockopt(zmq.IDENTITY, self.identity)
        c2s_socket.set_hwm(2)
        c2s_socket.connect(self.c2s)

        s2c_socket = context.socket(zmq.DEALER)
        s2c_socket.setsockopt(zmq.IDENTITY, self.identity)
        #s2c_socket.set_hwm(5)
        s2c_socket.connect(self.s2c)

        state = player.current_state()
        reward, isOver = 0, False
        while True:
            c2s_socket.send(dumps(
                (self.identity, state, reward, isOver)),
                copy=False)
            action = loads(s2c_socket.recv(copy=False))
            reward, isOver = player.action(action)
            state = player.current_state()

# compatibility
SimulatorProcess = StateExchangeSimulatorProcess

class SimulatorMaster(threading.Thread):
    """ A base thread to communicate with all StateExchangeSimulatorProcess.
        It should produce action for each simulator, as well as
        defining callbacks when a transition or an episode is finished.
    """
    __metaclass__ = ABCMeta

    class ClientState(object):
        def __init__(self):
            self.memory = []    # list of Experience

    class Experience(object):
        """ A transition of state, or experience"""
        def __init__(self, state, action, reward, **kwargs):
            """ kwargs: whatever other attribute you want to save"""
            self.state = state
            self.action = action
            self.reward = reward
            for k, v in six.iteritems(kwargs):
                setattr(self, k, v)

    def __init__(self, pipe_c2s, pipe_s2c):
        super(SimulatorMaster, self).__init__()
        self.daemon = True

        self.context = zmq.Context()

        self.c2s_socket = self.context.socket(zmq.PULL)
        self.c2s_socket.bind(pipe_c2s)
        self.c2s_socket.set_hwm(10)
        self.s2c_socket = self.context.socket(zmq.ROUTER)
        self.s2c_socket.bind(pipe_s2c)
        self.s2c_socket.set_hwm(10)

        # queueing messages to client
        self.send_queue = queue.Queue(maxsize=100)

        def f():
            msg = self.send_queue.get()
            # slow
            self.s2c_socket.send_multipart(msg, copy=False)
        self.send_thread = LoopThread(f)
        self.send_thread.daemon = True
        self.send_thread.start()

        # make sure socket get closed at the end
        def clean_context(soks, context):
            for s in soks:
                s.close()
            context.term()
        import atexit
        atexit.register(clean_context, [self.c2s_socket, self.s2c_socket], self.context)

    def run(self):
        self.clients = defaultdict(self.ClientState)
        #cnt = 0
        while True:
            #cnt += 1
            #if cnt % 3000 == 0:
                #print_total_timer()
            msg = loads(self.c2s_socket.recv(copy=False).bytes)
            ident, state, reward, isOver = msg
            client = self.clients[ident]

            # check if reward&isOver is valid
            # in the first message, only state is valid
            if len(client.memory) > 0:
                client.memory[-1].reward = reward
                if isOver:
                    self._on_episode_over(ident)
                else:
                    self._on_datapoint(ident)
            # feed state and return action
            self._on_state(state, ident)

    @abstractmethod
    def _on_state(self, state, ident):
        """response to state sent by ident. Preferrably an async call"""

    @abstractmethod
    def _on_episode_over(self, client):
        """ callback when the client just finished an episode.
            You may want to clear the client's memory in this callback.
        """

    def _on_datapoint(self, client):
        """ callback when the client just finished a transition
        """

    def __del__(self):
        self.socket.close()
        self.context.term()


class SimulatorProcessDF(SimulatorProcessBase):
    """ A simulator which contains a forward model itself, allowing
    it to produce data points directly """
    def __init__(self, idx, pipe_c2s):
        super(SimulatorProcessDF, self).__init__(idx)
        self.pipe_c2s = pipe_c2s

    def run(self):
        self.player = self._build_player()

        self.ctx = zmq.Context()
        self.c2s_socket = self.ctx.socket(zmq.PUSH)
        self.c2s_socket.setsockopt(zmq.IDENTITY, self.identity)
        self.c2s_socket.set_hwm(5)
        self.c2s_socket.connect(self.pipe_c2s)

        self._prepare()
        while True:
            dp = self._produce_datapoint()
            self.c2s_socket.send(dumps(
                (self.identity, dp)
                ), copy=False)

    @abstractmethod
    def _prepare(self):
        pass

    @abstractmethod
    def _produce_datapoint(self):
        pass


class SimulatorProcessSharedWeight(SimulatorProcessDF):
    """ A simulator process with an extra thread waiting for event,
    and take shared weight from shm.

    Start me under some CUDA_VISIBLE_DEVICES set!
    """
    def __init__(self, idx, pipe_c2s, evt, shared_dic):
        super(SimulatorProcessSharedWeight, self).__init__(idx, pipe_c2s)
        self.evt = evt
        self.shared_dic = shared_dic

    def _prepare(self):
        self._build_session()

        # start a thread to wait for evt
        def func():
            self.evt.wait()
            self._trigger_evt()
        self.evt_th = LoopThread(func, pausable=False)
        self.evt_th.start()

    @abstractmethod
    def _trigger_evt(self):
        pass
        #self.sess_updater.update(self.shared_dic['params'])

    @abstractmethod
    def _build_session(self):
        # build session and self.sess_updaer
        pass


if __name__ == '__main__':
    import random
    from tensorpack.RL import NaiveRLEnvironment
    class NaiveSimulator(SimulatorProcess):
        def _build_player(self):
            return NaiveRLEnvironment()
    class NaiveActioner(SimulatorActioner):
        def _get_action(self, state):
            time.sleep(1)
            return random.randint(1, 12)
        def _on_episode_over(self, client):
            #print("Over: ", client.memory)
            client.memory = []
            client.state = 0

    name = 'ipc://whatever'
    procs = [NaiveSimulator(k, name) for k in range(10)]
    [k.start() for k in procs]

    th = NaiveActioner(name)
    ensure_proc_terminate(procs)
    th.start()

    import time
    time.sleep(100)

